/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

//`define ISP_SUPPORT_SIN_SOUT

//
// Module Name: isp
//   ver wk11
// 
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//                                           no SIN/SOUT
// Number of Slice Flip Flops:           52      45
// Number of 4 input LUTs:               63      47
// Number of occupied Slices:            52      39
// Total Number of 4 input LUTs:         78      62
// Number of bonded IOBs:                39      39

// SIN       0aa1 aaaa  XXXX XXXX
//           XXXX XXXX  load_data
// 
// SOUT      1aa1 aaaa  store_data
//           XXXX XXXX  XXXX XXXX
// 
// SSTPR     0110 100a  PR[a]
//           XXXX XXXX  XXXX XXXX
// 
// SLD       0010 0I00  XXXX XXXX    ( if I == 1 , PR = PR +1 )
//           XXXX XXXX  load_data
// 
// SST       0110 0I00  store_data   ( if I == 1 , PR = PR +1 )
//           XXXX XXXX  XXXX XXXX

//              _______         _______         _______         _______
// SCK   ______|       |_______|       |_______|       |_______|       |
//
// MOSI <  MSB (bit7) ><  bit 6       > .............. <   bit 0       >
// MISO <  MSB (bit7) ><  bit 6       > .............. <   bit 0       >
// capt        @               @                               @
// r_count :
//  .          0       |       1       |    ....       |       7       | 
//  .          8       |       9       |    ....       |      15       | 
// load  r_count == 7 / posedge
// store r_count == 15 / posedge 
//
//
//              _______         _______         _______         _______
// CLK   ______|       |_______|       |_______|       |_______|       |
//
//                           ______________________________
// SCK   ___________________|                              |______________
//
//                     @               @              @                @
//                                   INPUT                          SHIFT
// r_count ==7 :
//              - - - - - - - - - - - - - - - - - - - 7 - - - - - - - -@ - 8
// r_load                               ______________
//                           __________|              |_______
// STORE_DATA                        {r_spi[6:0], r_spi_in } 
//
// r_count ==15:
//              - - - - - - - - - - - - - - - - - - -15 - - - - - - - -@ - 0
// r2_store                             ______________
//                           __________|              |_______
//
// r_data                                      @===========================
//                                                                     @
//
module isp (
		input CLK
		, input MOSI
		, input SCK
		, input CS
		, output MISO

		, output ISP_STORE
		, output ISP_LOAD
		, output [15:0] ISP_ADDR
		, output [7:0] ISP_STORE_DATA
		, input [7:0] ISP_LOAD_DATA
		);

    reg r_cs;
    reg r_sck;
    reg r_spi_in;
    reg [3:0]  r_count;
    reg [7:0]  r_spi;
    reg [15:0] r_pr;
    reg [7:0]  r_data;

    reg r_store;
    reg r2_store;
    reg r_load;
    reg r_sstpr_hi;
    reg r_sstpr_lo;
    reg r_pr_inc;
`ifdef ISP_SUPPORT_SIN_SOUT
    reg r_sin_sout;
    reg [5:0]  r_abs;
`endif
    assign ISP_STORE = r2_store;
    assign ISP_LOAD  = r_load;

    assign MISO = r_spi[7];
    assign ISP_STORE_DATA = {r_spi[6:0], r_spi_in };
`ifdef ISP_SUPPORT_SIN_SOUT
    assign ISP_ADDR = r_sin_sout ? { 10'h000, r_abs } : r_pr;
`else
    assign ISP_ADDR = r_pr;
`endif

`ifdef ISP_SUPPORT_SIN_SOUT
// SIN       0aa1 aaaa  XXXX XXXX
// SOUT      1aa1 aaaa  store_data
//  wire f_sin_sout = r_spi[4];
//  wire f_sin      = f_sin_sout & ~r_spi[7];
//  wire f_sout     = f_sin_sout &  r_spi[7];
    wire f7_sin_sout = r_spi[3];
    wire f7_sin      = f7_sin_sout & ~r_spi[6];
    wire f7_sout     = f7_sin_sout &  r_spi[6];
`endif
// SSTPR     0110 100a  PR[a]
//  wire f_sstpr    = r_spi[7:1] == 7'b0110100;
    wire f7_sstpr    = r_spi[6:0] == 7'b0110100;
// SLD       0010 0I00  XXXX XXXX    ( if I == 1 , PR = PR +1 )
// SST       0110 0I00  store_data   ( if I == 1 , PR = PR +1 )
//           * ** * **
//  wire f_sld_sst    = ~r_spi[7] & (r_spi[5:3] == 3'b100)
//			 & ~r_spi[1] & ~r_spi[0];
//  wire f_sld        = f_sld_sst & ~r_spi[6];
//  wire f_sst        = f_sld_sst &  r_spi[6];
//  wire f_pr_inc     = f_sld_sst &  r_spi[2];

    wire f7_sld_sst    = ~r_spi[6] & (r_spi[4:2] == 3'b100)
			 & ~r_spi[0] & ~MOSI;
    wire f7_sld        = f7_sld_sst & ~r_spi[5];
    wire f7_sst        = f7_sld_sst &  r_spi[5];
    wire f7_pr_inc     = f7_sld_sst &  r_spi[1];

`ifdef ISP_SUPPORT_SIN_SOUT
    wire f7_store = f7_sout | f7_sst;
    wire f7_load  = f7_sin  | f7_sld;

//    wire v_abs = { 10'h000, r_spi[6:5], r_spi[3:0] }
    wire [5:0] v7_abs = { r_spi[5:4], r_spi[2:0] , MOSI };
`else
    wire f7_store = f7_sst;
    wire f7_load  = f7_sld;
`endif

    always @(negedge CLK)
    begin
	r_cs        <= CS;
	r_sck       <= SCK;
	r_load   <= ~(~r_cs & CS) & (~r_sck & SCK) & (r_count == 7) & f7_load;
	r2_store <= ~(~r_cs & CS) & (~r_sck & SCK) & (r_count ==15) & r_store;

	if (~r_cs & CS)
	    begin
		r_count    <= 0;
		r_spi      <= 0;
	    end
	else if (~r_sck & SCK) // posedge SCK : INPUT
	    begin
		r_spi_in <= MOSI;
		if (r_count == 7)
	    	    begin
			r_store       <= f7_store;
			r_sstpr_hi    <= f7_sstpr &  MOSI;
			r_sstpr_lo    <= f7_sstpr & ~MOSI;
			r_pr_inc      <= f7_pr_inc;
`ifdef ISP_SUPPORT_SIN_SOUT
			r_sin_sout    <= f7_sin_sout;
			r_abs         <= v7_abs;
`endif
	    	    end
	    end
	else if (r_sck & ~SCK) // negedge SCK : SHIFT
	    begin
		r_count <= r_count + 1;
		if (r_count == 7) r_spi <= r_data;
		else r_spi <= {r_spi[6:0], r_spi_in };
		if (r_count == 15)
		  if (r_pr_inc) r_pr <= r_pr + 16'h0001; 
		  else
		    begin
			if (r_sstpr_hi) r_pr[15:8] <= {r_spi[6:0], r_spi_in };
			if (r_sstpr_lo) r_pr[ 7:0] <= {r_spi[6:0], r_spi_in };
		    end
	    end
    end

    always @(posedge CLK)
    begin
	if ( r_load ) r_data <= ISP_LOAD_DATA;
    end


`ifdef ISP_INITCLR
`else
    initial
    begin
	r_pr        <= 0;
	r_cs        <= 0;
	r_spi_in    <= 0;
	r_sck       <= 0;
	r_store     <= 0;
	r2_store    <= 0;
	r_load      <= 0;
	r_sstpr_hi  <= 0;
	r_sstpr_lo  <= 0;
	r_pr_inc    <= 0;
	r_data      <= 0;
`ifdef ISP_SUPPORT_SIN_SOUT
	r_sin_sout  <= 0;
	r_abs       <= 0;
`endif
    end
`endif

endmodule
